package MOAParser;

import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.TimeZone;

public class Location {
  public static final float CM2_IN_G = 980.6649999788F;

  // Device Status
  public static final byte GOOD_GPS = 0b00000001;
  public static final byte GOOD_COMM = 0b00000010;
  public static final byte GOOD_POWER = 0b00000100;
  public static final byte OLD_FIX = 0b00001000;
  private static final SimpleDateFormat simpleDate = new SimpleDateFormat("MM/dd HH:mm:ss");
  static {
    Location.simpleDate.setTimeZone(TimeZone.getTimeZone("UTC"));
  }
  private static final DecimalFormat simpleLat = new DecimalFormat("' '#00.0000000;-#00.0000000");
  private static final DecimalFormat oneDecimal = new DecimalFormat("#0.0");
  private static final DecimalFormat threeDecimal = new DecimalFormat("#0.000");

  // Real Vehicle Events
  public static final byte UNKNOWN = -1;
  public static final byte DOOR_UNLOCK = 0;
  public static final byte HEADING_CHANGE = 1;
  public static final byte IGNITION_OFF = 2;
  public static final byte ALIVE_HEARTBEAT = 3;
  public static final byte IGNITION_ON = 4;
  public static final byte MOVE_HEARTBEAT = 5;
  public static final byte INPUT_1_HIGH = 6;
  public static final byte INPUT_1_LOW = 7;
  public static final byte INPUT_2_HIGH = 8;
  public static final byte INPUT_2_LOW = 9;
  public static final byte INPUT_3_HIGH = 10;
  public static final byte INPUT_3_LOW = 11;
  public static final byte INPUT_4_HIGH = 12;
  public static final byte INPUT_4_LOW = 13;
  public static final byte INPUT_5_HIGH = 14;
  public static final byte INPUT_5_LOW = 15;
  public static final byte INPUT_6_HIGH = 16;
  public static final byte INPUT_6_LOW = 17;
  public static final byte INPUT_7_HIGH = 18;
  public static final byte INPUT_7_LOW = 19;
  public static final byte INPUT_8_HIGH = 20;
  public static final byte INPUT_8_LOW = 21;
  public static final byte POWER_CONNECTED = 22;
  public static final byte POWER_DISCONNECTED = 23;
  public static final byte STARTER_DISABLED = 24;
  public static final byte STARTER_ENABLED = 25;
  public static final byte STOP = 26;
  public static final byte MOVE_MINOR = 27;
  public static final byte VIRTUAL_IGN_ON = 28;
  public static final byte VIRTUAL_IGN_OFF = 29;
  public static final byte GPS_ACQUIRED = 30;
  public static final byte DOOR_LOCK = 31;
  public static final byte CRASH = 32;
  public static final byte HARD_ACCELERATION = 33;
  public static final byte HARD_BRAKING = 34;
  public static final byte HARD_LEFT = 35;
  public static final byte HARD_RIGHT = 36;
  public static final byte SOS_ALERT = 38;
  public static final byte COLD_BOOT = 39;
  public static final byte USER_LOCATE = 40;
  public static final byte STATUS_CHECK = 41;
  public static final byte WARM_BOOT = 42;
  public static final byte SHOCK = 43;
  public static final byte OBD = 45;
  public static final byte LOW_POWER = 52;
  public static final byte GPS_LOST = 101;
  public static final byte COMM_ACQUIRED = 102;
  public static final byte COMM_LOST = 103;
  public static final byte VIBRATION_WAKE = 104;
  public static final byte VOLTAGE_WAKE = 105;
  public static final byte ENTER_SLEEP = 107;
  public static final byte PERIODIC_BOOT = 108;
  public static final byte UPDATE_BOOT = 109;
  public static final byte COMM_RESET_UNAVAILABLE = 110;
  public static final byte GPS_RESET = 111;
  public static final byte COMM_RESET_NO_ACK = 112;
  public static final byte COMM_RESET_NO_AT = 113;
  public static final byte COMM_RESET_BY_CMD = 114;
  public static final byte ABSOLUTE_G_WAKE = 116;
  public static final byte DTC = 117;
  public static final byte TOW = 118;
  public static final byte SPEEDING = 119;
  public static final byte ENTER_POWER_PROTECT = (byte) 120;

  public static final byte CEL_STATUS_UNKNOWN = 0;
  public static final byte CEL_STATUS_ON = 1;
  public static final byte CEL_STATUS_OFF = 2;

  public static final int TYPE_TOWER = 1;
  public static final int TYPE_GPS = 2;

  private int date;       // Unix time in seconds
  private int distance;   // distance in meters
  private int engineSecs; // Cumulative total seconds engine has been on
  private byte rssi;      // Relative Received Signal Strength for the modem in dbm
  private byte inputs;    // Current input status (use byte masks for indiidual inputs)
  private int lat;        // Latitude * 10000000
  private int lng;        // Longitude * 10000000
  private short speed;    // Speed in kph
  private short accel;    // Acceleration in cm2
  private byte typeId;    // The location type (e.g. MOVE_HEARTBEAT)
  private int voltage;    // Voltage * 1000
  private String address = "";  // Address. Propagated when required
  private short heading;  // Heading (0 degrees to 360 degrees)
  private byte sats;      // Number of satellites currently acquired by GPS
  private short hdop;     // Horizontal Dilution of Precision of GPS position
  private int hac;        // Horizontal Accuracy of GPS position * 10
  private short offset;
  private int battVolts;  // Battery Voltage * 1000
  private byte battRem;   // level of battery charge remaining on device, 0 = none, 127 = full
  private byte celStatus;    // Check Engine Light (Unknown 0, on 1, off 2)
  private List<String> dtcCodes = new ArrayList<String>();
  private int registrationTime;        // time the device took to send registration message in seconds
  private int registrationTimeOutCount; // counter of how many times device has timed out when trying to connect to network while sending registration message
  private int temperature;              // device temperature in Celsius
  private byte msgType;

  // OBDII specific Types
  private String vin;     // Vehicle Identification Number
  private int obdType;
  private long vehOdometer;
  private long engineRuntime;
  private int fuelType;
  private long pid00;
  private long pid20;
  private long pid40;
  private long pid60;
  private long pid80;
  private long pidA0;

  public Location() {
  }

  public int getDate() {
    return date;
  }

  public int getDistance() {
    return distance;
  }

  public int getEngineSecs() {
    return engineSecs;
  }

  public byte getInputs() {
    return inputs;
  }

  public int getLat() {
    return lat;
  }

  public int getLng() {
    return lng;
  }

  public short getSpeed() {
    return speed;
  }

  public byte getTypeId() {
    return typeId;
  }

  public int getVoltage() {
    return voltage;
  }

  public void setDate(int value) {
    date = value;
  }

  public void setDistance(int value) {
    distance = value;
  }

  public void setEngineSecs(int value) {
    engineSecs = value;
  }

  public void setInputs(byte value) {
    inputs = value;
  }

  public void setLat(int value) {
    lat = value;
  }

  public void setLng(int value) {
    lng = value;
  }

  public void setSpeed(short value) {
    speed = value;
  }

  public void setTypeId(byte value) {
    typeId = value;
  }

  public void setVoltage(int value) {
    voltage = value;
  }

  public byte getRssi() {
    return rssi;
  }

  public void setRssi(byte rssi) {
    this.rssi = rssi;
  }

  public byte getStatus() {
    byte status = Location.GOOD_COMM;
    if (rssi > -90) {
      status = 0;
    }
    //Esky and Calamp and Xirgo devices
    if (hdop > 0 && sats > 4 && hdop < 15) {
      status = (byte) (status | Location.GOOD_GPS);

      //Starling devices only look at hac
    } else if (hac > 0 && sats > 4 && hac < 100) {
      status = (byte) (status | Location.GOOD_GPS);
    }
    return status;
  }

  public String getAddress() {
    return address;
  }

  public void setAddress(String address) {
    this.address = address;
  }

  public short getHeading() {
    return heading;
  }

  public void setHeading(short heading) {
    this.heading = heading;
  }

  public final static String getTypeString(byte id) {
    if (id == Location.DOOR_UNLOCK) {
      return "DOOR OPEN";
    } else if (id == Location.HEADING_CHANGE) {
      return "HEADING  ";
    } else if (id == Location.IGNITION_OFF) {
      return "IGN OFF  ";
    } else if (id == Location.ALIVE_HEARTBEAT) {
      return "HEARTBEAT";
    } else if (id == Location.IGNITION_ON) {
      return "IGN ON   ";
    } else if (id == Location.MOVE_HEARTBEAT) {
      return "MOVE     ";
    } else if (id == Location.INPUT_1_HIGH) {
      return "IN1 HIGH ";
    } else if (id == Location.INPUT_1_LOW) {
      return "IN1 LOW  ";
    } else if (id == Location.INPUT_2_HIGH) {
      return "IN2 HIGH ";
    } else if (id == Location.INPUT_2_LOW) {
      return "IN2 LOW  ";
    } else if (id == Location.INPUT_3_HIGH) {
      return "IN3 HIGH ";
    } else if (id == Location.INPUT_3_LOW) {
      return "IN3 LOW  ";
    } else if (id == Location.INPUT_4_HIGH) {
      return "IN4 HIGH ";
    } else if (id == Location.INPUT_4_LOW) {
      return "IN4 LOW  ";
    } else if (id == Location.INPUT_5_HIGH) {
      return "IN5 HIGH ";
    } else if (id == Location.INPUT_5_LOW) {
      return "IN5 LOW  ";
    } else if (id == Location.INPUT_6_HIGH) {
      return "IN6 HIGH ";
    } else if (id == Location.INPUT_6_LOW) {
      return "IN6 LOW  ";
    } else if (id == Location.INPUT_7_HIGH) {
      return "IN7 HIGH ";
    } else if (id == Location.INPUT_7_LOW) {
      return "IN7 LOW  ";
    } else if (id == Location.INPUT_8_HIGH) {
      return "IN8 HIGH ";
    } else if (id == Location.INPUT_8_LOW) {
      return "IN8 LOW  ";
    } else if (id == Location.POWER_CONNECTED) {
      return "PWR ON   ";
    } else if (id == Location.POWER_DISCONNECTED) {
      return "PWR OFF  ";
    } else if (id == Location.STARTER_DISABLED) {
      return "START OFF";
    } else if (id == Location.STARTER_ENABLED) {
      return "START ON ";
    } else if (id == Location.STOP) {
      return "STOP     ";
    } else if (id == Location.MOVE_MINOR) {
      return "MINOR    ";
    } else if (id == Location.VIRTUAL_IGN_ON) {
      return "VIGN ON  ";
    } else if (id == Location.VIRTUAL_IGN_OFF) {
      return "VIGN OFF ";
    } else if (id == Location.GPS_ACQUIRED) {
      return "GPS_ACQR ";
    } else if (id == Location.DOOR_LOCK) {
      return "DOOR LOCK";
    } else if (id == Location.CRASH) {
      return "CRASH    ";
    } else if (id == Location.HARD_ACCELERATION) {
      return "HRD ACCEL";
    } else if (id == Location.HARD_BRAKING) {
      return "HRD BRAKE";
    } else if (id == Location.HARD_LEFT) {
      return "HRD LEFT ";
    } else if (id == Location.HARD_RIGHT) {
      return "HRD RGHT ";
    } else if (id == Location.SPEEDING) {
      return "SPEEDING ";
    } else if (id == Location.SOS_ALERT) {
      return "SOS ALERT";
    } else if (id == Location.COLD_BOOT) {
      return "COLD BOOT";
    } else if (id == Location.WARM_BOOT) {
      return "WARM BOOT";
    } else if (id == Location.USER_LOCATE) {
      return "USER LOC ";
    } else if (id == Location.STATUS_CHECK) {
      return "STATUS CK";
    } else if (id == Location.SHOCK) {
      return "SHOCK    ";
    } else if (id == Location.OBD) {
      return "OBD      ";
    } else if (id == Location.LOW_POWER) {
      return "LOW POWER";
    } else if (id == Location.GPS_LOST) {
      return "GPS LOST ";
    } else if (id == Location.COMM_ACQUIRED) {
      return "COMM ACQR";
    } else if (id == Location.COMM_LOST) {
      return "COMM LOST";
    } else if (id == Location.VIBRATION_WAKE) {
      return "VIBR WAKE";
    } else if (id == Location.VOLTAGE_WAKE) {
      return "VOLT WAKE";
    } else if (id == Location.ENTER_SLEEP) {
      return "SLEEP    ";
    } else if (id == Location.PERIODIC_BOOT) {
      return "INTV BOOT";
    } else if (id == Location.UPDATE_BOOT) {
      return "UPDT BOOT";
    } else if (id == Location.COMM_RESET_UNAVAILABLE) {
      return "COM RESET NO AVAIL";
    } else if (id == Location.GPS_RESET) {
      return "GPS RESET";
    } else if (id == Location.COMM_RESET_NO_ACK) {
      return "COM RESET NO ACK";
    } else if (id == Location.COMM_RESET_NO_AT) {
      return "COM RESET NO AT";
    } else if (id == Location.COMM_RESET_BY_CMD) {
      return "COM RESET BY CMD";
    } else if (id == Location.ABSOLUTE_G_WAKE) {
      return "ABSOLUTEG";
    } else if (id == Location.DTC) {
      return "DTC      ";
    } else if (id == Location.TOW) {
      return "TOW      ";
    } else if (id == Location.ENTER_POWER_PROTECT) {
      return "ENTER PWR PROTECT";
    } else {
      int unsignedByte = id & 0x00ff;
      return "UNKNOWN: " + unsignedByte + " ";
    }
  }

  public void toString(StringBuilder out) {
    try {
      out.append(Location.simpleDate.format(date * 1000L));
    } catch (Exception inner) {
      out.append("Problem parsing date::" + date + " set to current server");
      date = Math.round(System.currentTimeMillis() / 1000);
      out.append(Location.simpleDate.format(date * 1000L));
    }
    if (offset > 0) {
      out.append(" off");
      out.append(offset * offset);
    }
    out.append(" ");
    int unsignedByte = typeId & 0x00ff;
    if (unsignedByte == OBD) {
      // Custom toString for OBDII type
      // If its an OBDII message, we shouldn't have any other locations, so
      // just print on new lines
      out.append("\n");
      out.append(String.format("Message Type: (%s) %s\n", unsignedByte, getTypeString(typeId)));
      out.append(String.format("rssi: %s\n", getRssi()));
      out.append(String.format("vin: %s\n", getVin()));
      out.append(String.format("obdType: %s\n", getObdType()));
      out.append(String.format("milStatus: %s\n", getCelStatusString()));
      out.append("dtcData: ");
      for (String dtcCode : getDtcCodes()) {
        out.append(String.format("%s, ", dtcCode));
      }
      out.append("\n");
      out.append(String.format("vehOdometer: %s\n", getVehOdometer()));
      out.append(String.format("engineRuntime: %s\n", getEngineRuntime()));
      out.append(String.format("fuelType: %s\n", getFuelType()));
      out.append(String.format("pid00: %s\n", getPid00()));
      out.append(String.format("pid20: %s\n", getPid20()));
      out.append(String.format("pid40: %s\n", getPid40()));
      out.append(String.format("pid60: %s\n", getPid60()));
      out.append(String.format("pid80: %s\n", getPid80()));
      out.append(String.format("pidA0: %s\n", getPidA0()));

    } else {
      out.append(Location.getTypeString(typeId) + " (" + String.format("%2s", unsignedByte) + ") ");
      out.append(" ");
      out.append(Location.simpleLat.format(lat / 10000000D));
      out.append(" ");
      out.append(String.format("%12s", Location.simpleLat.format(lng / 10000000D)));
      if (typeId != Location.MOVE_MINOR) {
        out.append(" s");
        out.append(sats);
      }
      if (hac > 0) {
        out.append(" hac");
        out.append(Location.oneDecimal.format(hac / 10D));
      } else if (hdop > 0) {
        out.append(" dop");
        out.append(hdop);
      }

      out.append(" ");
      if (rssi != 0) {
        out.append("r");
        out.append(rssi);
        out.append(" ");
      }
      out.append(String.valueOf(speed));

      out.append("kph");
      // out.append(" engs=");
      // out.append(engineSecs);
      if (typeId != Location.MOVE_MINOR) {
        out.append(" ");
        out.append(Integer.toBinaryString(inputs & 0xFF));
      }
      if (accel > 0) {
        out.append(" ");
        out.append(Location.threeDecimal.format(accel / CM2_IN_G));
        out.append("g");
      }
      if (distance > 0) {
        out.append(" ");
        out.append(distance);
        out.append("m");
      }
      if (voltage > 0) {
        out.append(" ");
        out.append(voltage / 1000D);
        out.append("V");
      }
      if (battVolts > 0) {
        out.append(" ");
        out.append(battVolts / 1000D);
        out.append("BV");
      }

      if (registrationTime > 0) {
        out.append(" ");
        out.append(registrationTime);
        out.append("s");
      }

      if (registrationTimeOutCount > 0) {
        out.append(" ");
        out.append(registrationTimeOutCount);
        out.append("rc");
      }

      if (temperature > 0) {
        out.append(" ");
        out.append(temperature);
        out.append("dc");
      }
    }
  }

  public String toString() {
    StringBuilder out = new StringBuilder(200);
    toString(out);
    return out.toString();
  }

  public short getAccel() {
    return accel;
  }

  public void setAccel(short accel) {
    this.accel = accel;
  }

  public void clone(Location other) {
    if (other != null) {
      setDate(other.getDate());
      setDistance(other.getDistance());
      setEngineSecs(other.getEngineSecs());
      setRssi(other.getRssi());
      setInputs(other.getInputs());
      setLat(other.getLat());
      setLng(other.getLng());
      setSpeed(other.getSpeed());
      setAccel(other.getAccel());
      setTypeId(other.getTypeId());
      setVoltage(other.getVoltage());
      setBattVolts(other.getBattVolts());
      setBattRem(other.getBattRem());
      setHeading(other.getHeading());
      setSats(other.getSats());
      setHdop(other.getHdop());
      setHac(other.getHac());
      setOffset(other.getOffset());
      setDtcCodes(other.getDtcCodes());
      setCelStatus(other.getCelStatus());
      setTemperature(other.getTemperature());
      setRegistrationTime(other.getRegistrationTime());
      setRegistrationTimeOutCount(other.getRegistrationTimeOutCount());
      setMsgType(other.getMsgType());
    }

  }

  public byte getSats() {
    return sats;
  }


  public void setSats(byte sats) {
    this.sats = sats;
  }

  public short getHdop() {
    return hdop;
  }

  public void setHdop(short hdop) {
    if (hdop > 255 || hdop < 0) {
      this.hdop = 255;
    } else {
      this.hdop = hdop;
    }
  }

  public int getHac() {
    return hac;
  }

  public void setHac(int hac) {
    if (hac < 0) {
      this.hac = 0;
    } else {
      this.hac = hac;
    }
  }

  public byte getBattRem() {
    return battRem;
  }

  public void setBattRem(byte value) {
    if (value < 0) {
      value = 0;
    }
    battRem = value;
  }

  public short getOffset() {
    return offset;
  }

  public void setOffset(short offset) {
    this.offset = offset;
  }

  public int getBattVolts() {
    return battVolts;
  }

  public void setBattVolts(int battVolts) {
    this.battVolts = battVolts;
  }

  public byte getCelStatus() {
    return celStatus;
  }

  public String getCelStatusString() {
    switch(celStatus) {
      case CEL_STATUS_UNKNOWN: 
        return "Unknown";
      case CEL_STATUS_ON: 
        return "On";
      case CEL_STATUS_OFF: 
        return "Off";
      default:
        return "Unknown code";
    }
  }

  public void setCelStatus(byte celStatus) {
    this.celStatus = celStatus;
  }

  /**
   * Gets the list of DTC codes
   * @return
   */
  public List<String> getDtcCodes() {
    return dtcCodes;
  }

  /**
   * Replaces the list of DTC codes
   * @param dtcCodes
   */
  public void setDtcCodes(List<String> dtcCodes) {
    this.dtcCodes = dtcCodes;
  }

  /**
   * Adds a DTC code to the list of dtcCodes.  Duplicates
   * are ignored
   * @param dtcCode
   */
  public void addDtcCode(String dtcCode) {
    if (findDtcCode(dtcCode) != null) {
      return;
    }
    dtcCodes.add(dtcCode);
  }

  /**
   * Considering that the backing of the dtcCodes are a list of strings,
   * this method sounds superfluous.   But it is a convenience method
   * to check if the given dtc code exists in the current list.
   * @param dtcCode
   * @return - the dtcCode, null if not found
   */
  public String findDtcCode(String dtcCode) {
    for (String dtc: dtcCodes) {
      if (dtc.equals(dtcCode)) {
        return dtc;
      }
    }
    return null;
  }

  public int getRegistrationTime() {
    return registrationTime;
  }

  public void setRegistrationTime(int registrationTime) {
    this.registrationTime = registrationTime;
  }

  public int getRegistrationTimeOutCount() {
    return registrationTimeOutCount;
  }

  public void setRegistrationTimeOutCount(int registrationTimeOutCount) {
    this.registrationTimeOutCount = registrationTimeOutCount;
  }

  public int getTemperature() {
    return temperature;
  }

  public void setTemperature(int temperature) {
    this.temperature = temperature;
  }

  public byte getMsgType() {
    return msgType;
  }

  public void setMsgType(byte msgType) {
    this.msgType = msgType;
  }

  public String getVin() {
    return vin;
  }

  public void setVin(String vin) {
    this.vin = vin;
  }

  public int getObdType() {
    return obdType;
  }

  public void setObdType(int obdType) {
    this.obdType = obdType;
  }

  public long getVehOdometer() {
    return vehOdometer;
  }

  public void setVehOdometer(long vehOdometer) {
    this.vehOdometer = vehOdometer;
  }

  public long getEngineRuntime() {
    return engineRuntime;
  }

  public void setEngineRuntime(long engineRuntime) {
    this.engineRuntime = engineRuntime;
  }

  public int getFuelType() {
    return fuelType;
  }

  public void setFuelType(int fuelType) {
    this.fuelType = fuelType;
  }

  public long getPid00() {
    return pid00;
  }

  public void setPid00(long pid00) {
    this.pid00 = pid00;
  }

  public long getPid20() {
    return pid20;
  }

  public void setPid20(long pid20) {
    this.pid20 = pid20;
  }

  public long getPid40() {
    return pid40;
  }

  public void setPid40(long pid40) {
    this.pid40 = pid40;
  }

  public long getPid60() {
    return pid60;
  }

  public void setPid60(long pid60) {
    this.pid60 = pid60;
  }

  public long getPid80() {
    return pid80;
  }

  public void setPid80(long pid80) {
    this.pid80 = pid80;
  }

  public long getPidA0() {
    return pidA0;
  }

  public void setPidA0(long pidA0) {
    this.pidA0 = pidA0;
  }

}
