package MOAParser;

import java.nio.ByteBuffer;
import java.nio.BufferUnderflowException;
import java.nio.BufferOverflowException;
import java.nio.ReadOnlyBufferException;

public class MoaParser {

  public static final int SERVER_LOC_ACK = 42;
  public static final int SERVER_LOC_NAK = 43;

  public static final int MESSAGE_TYPE_EVENT = 1;
  public static final int MESSAGE_TYPE_OBDII = 45;
  public static final int MESSAGE_TYPE_CUSTOM = 30;
  public static final int MESSAGE_TYPE_CMD_RESPONSE = 8;

  // Constants from Location
  public static final byte CEL_STATUS_UNKNOWN = 0;
  public static final byte CEL_STATUS_ON = 1;
  public static final byte CEL_STATUS_OFF = 2;
  public static final byte MOVE_MINOR = 27;
  public static final byte OBD = 45;
  public static final byte TOW = 118;

  // Constants from AppData
  public static final float CM2_IN_G = 980.6649999788F;

  /**
   * Pass in a ByteBuffer containing the raw data to be parsed
   * The parsedLocation will be in IParsedLocation
   * A ByteBuffer containing the ACK or NACK to be sent will be returned
   * or null if no ACK required
   *
   */
  public ByteBuffer handleLocation(ByteBuffer rawInput, IParsedLocation parsed)  {

    parsed.setTypeId(IParsedLocation.IMEI);
    ByteBufferReader dis = new ByteBufferReader(rawInput);
    ByteBufferWriter ackWriter = new ByteBufferWriter(100);
    long imei = dis.getLongBE();
    parsed.setDeviceId("" + imei);
    int messageType = dis.getUnsignedByte();
    int simpleTypeId = messageType & 0x3f; // Bit 7=Use an ack, Bit 6=use bitmap
    boolean needAck = (messageType & 0x0080) != 0;
    int checkBitMap = messageType & 0x0040;
    boolean hasBitmap = checkBitMap != 0;

    if (simpleTypeId == MoaParser.MESSAGE_TYPE_CMD_RESPONSE) {
      parsed.setStatus(IParsedLocation.STATUS_CUSTOM);
      // print command response out into the logs
      boolean cmdResponse = dis.getIntBE() == 1;
      int cmdLength = dis.getIntBE();
      String cmd = dis.getString(cmdLength);
      int responseLength = dis.getIntBE();
      String response = dis.getString(responseLength);
      System.out.println("Ignore[181]: CMD Response: " +  (cmdResponse ? " PASS:" : " FAIL:") + 
          cmd.toString() + " -> " + response.toString());
      parsed.setStatus(IParsedLocation.STATUS_BAD);

    } else if (simpleTypeId == MoaParser.MESSAGE_TYPE_OBDII) {

      ObdiiPayload.parse(dis, parsed);

    } else if (simpleTypeId == MoaParser.MESSAGE_TYPE_EVENT || simpleTypeId == MoaParser.MESSAGE_TYPE_CUSTOM) {

      try {
        long mainBitmap = 0;
        long minorBitmap = 0;
        long extendedBitmap = 0;
        mainBitmap = (dis.getUnsignedByte() << 16)
            + (dis.getUnsignedByte() << 8)
            + dis.getUnsignedByte();

        // HERE
        EventPayload.parse(dis, parsed, mainBitmap, messageType, hasBitmap);

        if ((mainBitmap & EventPayload.MINOR_MAP) > 0) {
          minorBitmap = dis.getUnsignedByte();
        }
        if ((mainBitmap & EventPayload.EXTENDED_MAP) > 0) {
          extendedBitmap = dis.getUnsignedIntBE();
        }
        if ((mainBitmap & EventPayload.EXTENDED_MAP) > 0) {
          //HERE
          ExtendedPayload.parse(dis, parsed, extendedBitmap);
        }
        if ((mainBitmap & EventPayload.MINOR_MAP) > 0) {
          //HERE
          MinorPayload.parse(dis, parsed, minorBitmap);
        }

        if (needAck) {
          ackWriter.clear();
          ackWriter.putByte(SERVER_LOC_ACK);
          ackWriter.putByte(parsed.getSequence());
        }
      } catch (Exception e) {
        e.printStackTrace();
        System.out.println("Ignore[101]: Malformed msg:::" + e.getLocalizedMessage());
        // send result which guarantees last record deletion
        ackWriter.clear();
        ackWriter.putByte(SERVER_LOC_NAK);
        parsed.setStatus(IParsedLocation.STATUS_BAD);
      }
    } else {
      System.out.println("Ignore[102]: Unknown event type=" + simpleTypeId);
      ackWriter.clear();
      ackWriter.putByte(SERVER_LOC_NAK);
      parsed.setStatus(IParsedLocation.STATUS_BAD);
    }
    // DUCK-739. Ignore TOW events for PUI
    if (simpleTypeId == TOW) {
      System.out.println("Ignore[103]: Tow event");
      parsed.setStatus(IParsedLocation.STATUS_BAD);
    }
    // Return null if no ack, otherwise return the byte buffer
    if (ackWriter.size() == 0) {
      return null;
    } else {
      return ackWriter.getByteBuffer();
    }
  }

}
